import abc

from itertools import cycle


class ProxyProvider(abc.ABC):
    @abc.abstractmethod
    def get_proxy(self):
        return NotImplemented


class NoProxyProvier(ProxyProvider):
    def get_proxy(self):
        return {}


class SingleProxyProvider(ProxyProvider):
    def __init__(self, proxy=None):
        self._proxy = proxy

    def get_proxy(self):
        return self._proxy


class RoundRobinProxiesProvider(ProxyProvider):
    def __init__(self, proxies: list):
        self._proxies = proxies
        self._proxies_cycle = cycle(proxies)

    @property
    def proxies(self):
        return self._proxies

    @proxies.setter
    def proxies(self, proxies: list):
        if not isinstance(proxies, list):
            raise ValueError("Proxies only accept list")

        self._proxies = proxies
        self._proxies_cycle = cycle(proxies)

    def get_proxy(self):
        return next(self._proxies_cycle)


_provider_instance = NoProxyProvier()


def reset_proxy_provider():
    configure_proxy_provider(NoProxyProvier())


def configure_proxy_provider(provider_instance):
    global _provider_instance
    if not isinstance(provider_instance, ProxyProvider):
        raise BaseException("proxy provider should be a ProxyProvider object")
    _provider_instance = provider_instance


def get_proxies():
    return _provider_instance.get_proxy()



# -*- coding: utf-8 -*-

import unittest
import twstock
from twstock import realtime


class RealtimeTest(unittest.TestCase):
    def test_realtime_field(self):
        self.assertCountEqual(
            realtime.get_raw("2330").keys(), twstock.mock.get_stock_info("2330").keys()
        )

    def test_realtime_get_raw(self):
        self.assertIn("msgArray", realtime.get_raw("2330"))

    def test_realtime_get_blank(self):
        stock = realtime.get("")

        self.assertFalse(stock["success"])
        self.assertIn("rtmessage", stock)
        self.assertIn("rtcode", stock)

    def test_realtime_get_bad_id(self):
        stock = realtime.get("9999")

        self.assertFalse(stock["success"])
        self.assertIn("rtmessage", stock)
        self.assertIn("rtcode", stock)

        stock = realtime.get(["9999", "8888"])

        self.assertFalse(stock["success"])
        self.assertIn("rtmessage", stock)
        self.assertIn("rtcode", stock)

    def test_realtime_get_tpex_id(self):
        stock = realtime.get("6223")

        self.assertTrue(stock["success"])
        self.assertEqual(stock["info"]["code"], "6223")


class MockRealtimeTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        realtime.mock = True

    @classmethod
    def tearDownClass(cls):
        realtime.mock = False

    def test_mock_one_stock_id(self):
        s = realtime.get("2330")

        self.assertTrue(s["success"])
        self.assertEqual(s["info"]["code"], "2330")
        self.assertEqual(s["realtime"]["latest_trade_price"], "-")
        self.assertEqual(
            s["realtime"]["best_bid_price"],
            ["848.0000", "847.0000", "846.0000", "845.0000", "844.0000"],
        )

    @unittest.skip("Dont want to fix this, is about the code in realtime")
    def test_mock_multiple_stock_id(self):
        s = realtime.get(["2330", "2337"])

        self.assertTrue(s["success"])
        self.assertCountEqual(s.keys(), ["2330", "2337", "success"])
        self.assertTrue(s["2330"]["success"])
        self.assertEqual(s["2330"]["info"]["code"], "2330")
        self.assertEqual(s["2330"]["realtime"]["latest_trade_price"], "214.50")
        self.assertTrue(s["2337"]["success"])



# -*- coding: utf-8 -*-

import datetime
import urllib.parse
from collections import namedtuple

from twstock.proxy import get_proxies

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

import requests

try:
    from . import analytics
    from .codes import codes
except ImportError as e:
    if e.name == "lxml":
        # Fix #69
        raise e
    import analytics
    from codes import codes


TWSE_BASE_URL = "http://www.twse.com.tw/"
TPEX_BASE_URL = "http://www.tpex.org.tw/"
DATATUPLE = namedtuple(
    "Data",
    [
        "date",
        "capacity",
        "turnover",
        "open",
        "high",
        "low",
        "close",
        "change",
        "transaction",
        "note",
    ],
)


class StockIDNotFoundError(KeyError):
    """Exception raised when a stock ID is not found in the database."""

    def __str__(self):
        return super().__str__().replace("\\n", "\n")


class BaseFetcher(object):
    def fetch(self, year, month, sid, retry):
        pass

    def _convert_date(self, date):
        """Convert '106/05/01' to '2017/05/01'"""
        return "/".join([str(int(date.split("/")[0]) + 1911)] + date.split("/")[1:])

    def _make_datatuple(self, data):
        pass

    def purify(self, original_data):
        pass


class TWSEFetcher(BaseFetcher):
    REPORT_URL = urllib.parse.urljoin(TWSE_BASE_URL, "exchangeReport/STOCK_DAY")

    def __init__(self):
        pass

    def fetch(self, year: int, month: int, sid: str, retry: int = 5):
        params = {"date": "%d%02d01" % (year, month), "stockNo": sid}
        for retry_i in range(retry):
            r = requests.get(self.REPORT_URL, params=params, proxies=get_proxies())
            try:
                data = r.json()
            except JSONDecodeError:
                continue
            else:
                break
        else:
            # Fail in all retries
            data = {"stat": "", "data": []}

        if data["stat"] == "OK":
            data["data"] = self.purify(data)
        else:
            data["data"] = []
        return data

    def _make_datatuple(self, data):
        data[0] = datetime.datetime.strptime(self._convert_date(data[0]), "%Y/%m/%d")
        data[1] = int(data[1].replace(",", ""))
        data[2] = int(data[2].replace(",", ""))
        data[3] = None if data[3] == "--" else float(data[3].replace(",", ""))
        data[4] = None if data[4] == "--" else float(data[4].replace(",", ""))
        data[5] = None if data[5] == "--" else float(data[5].replace(",", ""))
        data[6] = None if data[6] == "--" else float(data[6].replace(",", ""))
        # +/-/X表示漲/跌/不比價
        data[7] = float(
            0.0 if data[7].replace(",", "") == "X0.00" else data[7].replace(",", "")
        )
        data[8] = int(data[8].replace(",", ""))
        data[9] = data[9]
        return DATATUPLE(*data)

    def purify(self, original_data):
        return [self._make_datatuple(d) for d in original_data["data"]]


class TPEXFetcher(BaseFetcher):
    REPORT_URL = urllib.parse.urljoin(
        TPEX_BASE_URL, "www/zh-tw/afterTrading/tradingStock"
    )

    def __init__(self):
        pass

    def fetch(self, year: int, month: int, sid: str, retry: int = 5):
        params = {
            "date": "%d/%02d/01" % (year, month),
            "code": sid,
            "response": "json",
        }
        for retry_i in range(retry):
            r = requests.get(self.REPORT_URL, params=params, proxies=get_proxies())
            try:
                data = r.json()
            except JSONDecodeError:
                continue
            else:
                break
        else:
            # Fail in all retries
            data = {"aaData": []}

        data["data"] = self.purify(data)
        return data

    def _convert_date(self, date):
        """Convert '106/05/01' to '2017/05/01'"""
        return "/".join([str(int(date.split("/")[0]) + 1911)] + date.split("/")[1:])

    def _make_datatuple(self, data):
        data[0] = datetime.datetime.strptime(
            self._convert_date(data[0].replace("*", "")), "%Y/%m/%d"
        )
        data[1] = int(data[1].replace(",", "")) * 1000
        data[2] = int(data[2].replace(",", "")) * 1000
        data[3] = None if data[3] == "--" else float(data[3].replace(",", ""))
        data[4] = None if data[4] == "--" else float(data[4].replace(",", ""))
        data[5] = None if data[5] == "--" else float(data[5].replace(",", ""))
        data[6] = None if data[6] == "--" else float(data[6].replace(",", ""))
        data[7] = float(data[7].replace(",", ""))
        data[8] = int(data[8].replace(",", ""))
        return DATATUPLE(*data, note="")

    def purify(self, original_data):
        return [self._make_datatuple(d) for d in original_data["tables"][0]["data"]]


DATA_FETCHER = {
    "twse": TWSEFetcher,
    "tpex": TPEXFetcher,
}


class Stock(analytics.Analytics):
    def __init__(
        self,
        sid: str,
        initial_fetch: bool = True,
        force_data_source: str = None,
    ):
        """
        Args:
            sid (str): Stock ID
            initial_fetch (bool): Fetch data when initializing
            force_data_source (str): Force data source, can be 'twse' or 'tpex',
                if not set, will use the data source from stock codes database,
                if set, it also implied that we will not check if the stock ID is in the database.
        """

        if force_data_source is None and sid not in codes:
            raise StockIDNotFoundError(
                f'Stock ID "{sid}" not found in database.\n'
                "Either you misspelled it or it is not in the database.\n"
                "If you believe your stock ID is correct, please update stock codes database using:\n"
                "  - CLI: twstock -U\n"
                '  - Python: python -c "import twstock; twstock.__update_codes()"'
            )

        self.sid = sid
        self.fetcher = DATA_FETCHER.get(
            force_data_source if force_data_source else codes[sid].data_source
        )()
        self.raw_data = []
        self.data = []

        # Init data
        if initial_fetch:
            self.fetch_31()

    def _month_year_iter(self, start_month, start_year, end_month, end_year):
        ym_start = 12 * start_year + start_month - 1
        ym_end = 12 * end_year + end_month
        for ym in range(ym_start, ym_end):
            y, m = divmod(ym, 12)
            yield y, m + 1

    def fetch(self, year: int, month: int):
        """Fetch year month data"""
        self.raw_data = [self.fetcher.fetch(year, month, self.sid)]
        self.data = self.raw_data[0]["data"]
        return self.data

    def fetch_from(self, year: int, month: int):
        """Fetch data from year, month to current year month data"""
        self.raw_data = []
        self.data = []
        today = datetime.datetime.today()
        for year, month in self._month_year_iter(month, year, today.month, today.year):
            self.raw_data.append(self.fetcher.fetch(year, month, self.sid))
            self.data.extend(self.raw_data[-1]["data"])
        return self.data

    def fetch_31(self):
        """Fetch 31 days data"""
        today = datetime.datetime.today()
        before = today - datetime.timedelta(days=60)
        self.fetch_from(before.year, before.month)
        self.data = self.data[-31:]
        return self.data

    @property
    def date(self):
        return [d.date for d in self.data]

    @property
    def capacity(self):
        return [d.capacity for d in self.data]

    @property
    def turnover(self):
        return [d.turnover for d in self.data]

    @property
    def price(self):
        return [d.close for d in self.data]

    @property
    def high(self):
        return [d.high for d in self.data]

    @property
    def low(self):
        return [d.low for d in self.data]

    @property
    def open(self):
        return [d.open for d in self.data]

    @property
    def close(self):
        return [d.close for d in self.data]

    @property
    def change(self):
        return [d.change for d in self.data]

    @property
    def transaction(self):
        return [d.transaction for d in self.data]



# -*- coding: utf-8 -*-
#
# Usage: Download all stock code info from TWSE
#
# TWSE equities = 上市證券
# TPEx equities = 上櫃證券
#

import os
import csv
from collections import namedtuple

import requests
from lxml import etree

from twstock.proxy import get_proxies

TWSE_EQUITIES_URL = "https://isin.twse.com.tw/isin/C_public.jsp?strMode=2"
TPEX_EQUITIES_URL = "https://isin.twse.com.tw/isin/C_public.jsp?strMode=4"
ROW = namedtuple(
    "Row", ["type", "code", "name", "ISIN", "start", "market", "group", "CFI"]
)


def make_row_tuple(typ, row):
    code, name = row[1].split("\u3000")
    return ROW(typ, code, name, *row[2:-1])


def fetch_data(url):
    r = requests.get(url, proxies=get_proxies())
    root = etree.HTML(r.text)
    trs = root.xpath("//tr")[1:]

    result = []
    typ = ""
    for tr in trs:
        tr = list(map(lambda x: x.text, tr.iter()))
        if len(tr) == 4:
            # This is type
            typ = tr[2].strip(" ")
        else:
            # This is the row data
            result.append(make_row_tuple(typ, tr))
    return result


def to_csv(url, path):
    data = fetch_data(url)
    with open(path, "w", newline="", encoding="utf_8") as csvfile:
        writer = csv.writer(
            csvfile, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        writer.writerow(data[0]._fields)
        for d in data:
            writer.writerow([_ for _ in d])


def __update_codes():
    def get_directory():
        return os.path.dirname(os.path.abspath(__file__))

    to_csv(TWSE_EQUITIES_URL, os.path.join(get_directory(), "twse_equities.csv"))
    to_csv(TPEX_EQUITIES_URL, os.path.join(get_directory(), "tpex_equities.csv"))


if __name__ == "__main__":
    to_csv(TWSE_EQUITIES_URL, "twse_equities.csv")
    to_csv(TPEX_EQUITIES_URL, "tpex_equities.csv")



import unittest

from twstock.proxy import get_proxies, configure_proxy_provider, reset_proxy_provider
from twstock.proxy import SingleProxyProvider
from twstock.proxy import RoundRobinProxiesProvider


class ProxyProviderTest(unittest.TestCase):
    def setUp(self):
        reset_proxy_provider()

    def tearDown(self):
        reset_proxy_provider()

    def test_configure(self):
        # default values are empty
        self.assertDictEqual({}, get_proxies())

        # configure fake proxy
        configure_proxy_provider(
            SingleProxyProvider(dict(http="http-proxy", https="https-proxy"))
        )
        self.assertEqual("http-proxy", get_proxies()["http"])
        self.assertEqual("https-proxy", get_proxies()["https"])

        # reset proxy
        reset_proxy_provider()
        self.assertDictEqual({}, get_proxies())

    def test_rr_proxies_provider(self):
        proxies = ["a", "b", "c"]
        rr_provider = RoundRobinProxiesProvider(proxies)

        for _ in range(3):
            for p in proxies:
                self.assertEqual(rr_provider.get_proxy(), p)

        proxies = ["d", "e", "f"]
        rr_provider.proxies = proxies
        for _ in range(3):
            for p in proxies:
                self.assertEqual(rr_provider.get_proxy(), p)



# -*- coding: utf-8 -*-
#
# Usage: Load all Taiwan stock code info from csv file
#
# TWSE equities = 上市證券
# TPEx equities = 上櫃證券
#

import csv
import os
from collections import namedtuple


ROW = namedtuple(
    "StockCodeInfo",
    ["type", "code", "name", "ISIN", "start", "market", "group", "CFI", "data_source"],
)
PACKAGE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
TPEX_EQUITIES_CSV_PATH = os.path.join(PACKAGE_DIRECTORY, "tpex_equities.csv")
TWSE_EQUITIES_CSV_PATH = os.path.join(PACKAGE_DIRECTORY, "twse_equities.csv")

codes = {}
tpex = {}
twse = {}


def read_csv(path, types):
    global codes, twse, tpex
    with open(path, newline="", encoding="utf_8") as csvfile:
        reader = csv.reader(csvfile)
        csvfile.readline()
        for row in reader:
            row = ROW(*(item.strip() for item in row), data_source=types)
            codes[row.code] = row
            if types == "tpex":
                tpex[row.code] = row
            else:
                twse[row.code] = row


read_csv(TPEX_EQUITIES_CSV_PATH, "tpex")
read_csv(TWSE_EQUITIES_CSV_PATH, "twse")



# -*- coding: utf-8 -*-

import io
import sys
import twstock

# XXX: Repalce sys.stdout prevent Windows UnicodeEncodeError on cmd.exe
stdout = io.TextIOWrapper(
    getattr(sys.stdout, "buffer", sys.stdout), encoding="utf-8", errors="replace"
)


def run(argv):
    print("四大買賣點判斷 Best Four Point", file=stdout)
    print("------------------------------", file=stdout)
    for sid in argv:
        bfp = twstock.BestFourPoint(twstock.Stock(sid))
        bfp = bfp.best_four_point()
        print("%s: " % (sid), end="", file=stdout)
        if bfp:
            if bfp[0]:
                print("Buy  ", bfp[1], file=stdout)
            else:
                print("Sell ", bfp[1], file=stdout)
        else:
            print("Don't touch", file=stdout)



# -*- coding: utf-8 -*-


class LegacyAnalytics(object):
    """Legacy analytics from toomore/grs"""

    def cal_continue(self, list_data):
        """計算持續天數

        :rtype: int
        :returns: 向量數值：正數向上、負數向下。
        """
        diff_data = []
        for i in range(1, len(list_data)):
            if list_data[-i] > list_data[-i - 1]:
                diff_data.append(1)
            else:
                diff_data.append(-1)
        cont = 0
        for value in diff_data:
            if value == diff_data[0]:
                cont += 1
            else:
                break
        return cont * diff_data[0]

    def moving_average(self, data, days):
        """計算移動平均數

        :rtype: 序列 舊→新
        """
        result = []
        data = data[:]
        for dummy in range(len(data) - int(days) + 1):
            result.append(round(sum(data[-days:]) / days, 2))
            data.pop()
        result.reverse()
        return result

    def ma_bias_ratio(self, date1, date2, data):
        """計算乖離率（均價）
        date1 - date2

        :param int data1: n 日
        :param int data2: m 日
        :rtype: 序列 舊→新
        """
        data1 = self.moving_average(data, date1)
        data2 = self.moving_average(data, date2)
        cal_list = []
        for i in range(1, min(len(data1), len(data2)) + 1):
            cal_list.append(data1[-i] - data2[-i])
        cal_list.reverse()
        return cal_list

    def ma_bias_ratio_point(cls, data, sample=5, positive_or_negative=False):
        """判斷轉折點位置

        :param list data: 計算資料
        :param int sample: 計算的區間樣本數量
        :param bool positive_or_negative: 正乖離 為 True，負乖離 為 False
        :rtype: tuple
        :returns: (True or False, 第幾個轉折日, 轉折點值)
        """
        sample_data = data[-sample:]
        if positive_or_negative:  # 正
            ckvalue = max(sample_data)  # 尋找最大值
            preckvalue = max(sample_data) > 0  # 區間最大值必須為正
        else:
            ckvalue = min(sample_data)  # 尋找最小值
            preckvalue = max(sample_data) < 0  # 區間最大值必須為負
        return (
            sample - sample_data.index(ckvalue) < 4
            and sample_data.index(ckvalue) != sample - 1
            and preckvalue,
            sample - sample_data.index(ckvalue) - 1,
            ckvalue,
        )


class LegacyBestFourPoint(object):
    """四大買點組合

    :param grs.Stock data: 個股資料
    """

    def __init__(self, data):
        self.data = data

    def bias_ratio(self, position=False):
        """判斷乖離

        :param bool positive_or_negative: 正乖離 為 True，負乖離 為 False
        """
        return self.data.ma_bias_ratio_pivot(
            self.data.ma_bias_ratio(3, 6), position=position
        )

    def check_plus_bias_ratio(self):
        """正乖離扣至最大"""
        return self.bias_ratio(True)

    def check_mins_bias_ratio(self):
        """負乖離扣至最大"""
        return self.bias_ratio()

    ##### 四大買點 #####
    def best_buy_1(self):
        """量大收紅"""
        result = (
            self.data.capacity[-1] > self.data.capacity[-2]
            and self.data.price[-1] > self.data.open[-1]
        )
        return result

    def best_buy_2(self):
        """量縮價不跌"""
        result = (
            self.data.capacity[-1] < self.data.capacity[-2]
            and self.data.price[-1] > self.data.price[-2]
        )
        return result

    def best_buy_3(self):
        """三日均價由下往上"""
        return self.data.continuous(self.data.moving_average(self.data.price, 3)) == 1

    def best_buy_4(self):
        """三日均價大於六日均價"""
        return (
            self.data.moving_average(self.data.price, 3)[-1]
            > self.data.moving_average(self.data.price, 6)[-1]
        )

    ##### 四大賣點 #####
    def best_sell_1(self):
        """量大收黑"""
        result = (
            self.data.capacity[-1] > self.data.capacity[-2]
            and self.data.price[-1] < self.data.open[-1]
        )
        return result

    def best_sell_2(self):
        """量縮價跌"""
        result = (
            self.data.capacity[-1] < self.data.capacity[-2]
            and self.data.price[-1] < self.data.price[-2]
        )
        return result

    def best_sell_3(self):
        """三日均價由上往下"""
        return self.data.continuous(self.data.moving_average(self.data.price, 3)) == -1

    def best_sell_4(self):
        """三日均價小於六日均價"""
        return (
            self.data.moving_average(self.data.price, 3)[-1]
            < self.data.moving_average(self.data.price, 6)[-1]
        )

    def best_four_point_to_buy(self):
        """判斷是否為四大買點

        :rtype: str or False
        """
        result = []
        if self.check_mins_bias_ratio() and (
            self.best_buy_1()
            or self.best_buy_2()
            or self.best_buy_3()
            or self.best_buy_4()
        ):
            if self.best_buy_1():
                result.append(self.best_buy_1.__doc__.strip())
            if self.best_buy_2():
                result.append(self.best_buy_2.__doc__.strip())
            if self.best_buy_3():
                result.append(self.best_buy_3.__doc__.strip())
            if self.best_buy_4():
                result.append(self.best_buy_4.__doc__.strip())
            result = ", ".join(result)
        else:
            result = False
        return result

    def best_four_point_to_sell(self):
        """判斷是否為四大賣點

        :rtype: str or False
        """
        result = []
        if self.check_plus_bias_ratio() and (
            self.best_sell_1()
            or self.best_sell_2()
            or self.best_sell_3()
            or self.best_sell_4()
        ):
            if self.best_sell_1():
                result.append(self.best_sell_1.__doc__.strip())
            if self.best_sell_2():
                result.append(self.best_sell_2.__doc__.strip())
            if self.best_sell_3():
                result.append(self.best_sell_3.__doc__.strip())
            if self.best_sell_4():
                result.append(self.best_sell_4.__doc__.strip())
            result = ", ".join(result)
        else:
            result = False
        return result

    def best_four_point(self):
        """判斷買點或賣點

        :rtype: tuple
        :returns: (bool, str)
        """
        buy = self.best_four_point_to_buy()
        sell = self.best_four_point_to_sell()

        if buy:
            return True, buy
        elif sell:
            return False, sell

        return None



import unittest
from twstock import mock


class MockTest(unittest.TestCase):
    def test_mock_get_stock_info_will_work(self):
        self.assertIn("msgArray", mock.get_stock_info("2330"))

    def test_mock_get_stock_info_raw_data(self):
        self.assertCountEqual(
            mock.get_stock_info("2330").keys(),
            [
                "msgArray",
                "userDelay",
                "rtmessage",
                "referer",
                "queryTime",
                "rtcode",
                "exKey",
                "cachedAlive",
            ],
        )

    def test_mock_get_stock_info_msgarray(self):
        self.assertEqual(mock.get_stock_info("2330")["msgArray"][0]["c"], "2330")

    def test_mock_get_stock_info_will_change_in_different_index(self):
        self.assertNotEqual(
            mock.get_stock_info("2330", 0), mock.get_stock_info("2330", 1)
        )
        self.assertNotEqual(
            mock.get_stock_info("2330", 1), mock.get_stock_info("2330", 2)
        )
        self.assertNotEqual(
            mock.get_stock_info("2330", 0), mock.get_stock_info("2330", 2)
        )



# -*- coding: utf-8 -*-

import unittest
from twstock import cli


class CLIFunctionTest(unittest.TestCase):
    def setUp(self):
        self.stocks = ["2330", "6223"]

    def test_best_four_point(self):
        cli.best_four_point.run(self.stocks)

    def test_stock(self):
        cli.stock.run(self.stocks)



import datetime
import unittest
from twstock import stock


class FetcherTest(object):
    def test_convert_date(self):
        date = "106/05/01"
        cv_date = self.fetcher._convert_date(date)
        self.assertEqual(cv_date, "2017/05/01")

    def test_make_datatuple(self):
        data = [
            "106/05/02",
            "45,851,963",
            "9,053,856,108",
            "198.50",
            "199.00",
            "195.50",
            "196.50",
            "+2.00",
            "15,718",
        ]
        dt = self.fetcher._make_datatuple(data)
        self.assertEqual(dt.date, datetime.datetime(2017, 5, 2))
        self.assertEqual(dt.capacity, 45851963)
        self.assertEqual(dt.turnover, 9053856108)
        self.assertEqual(dt.open, 198.5)
        self.assertEqual(dt.high, 199.0)
        self.assertEqual(dt.low, 195.5)
        self.assertEqual(dt.close, 196.5)
        self.assertEqual(dt.change, 2.0)
        self.assertEqual(dt.transaction, 15718)

    def test_make_datatuple_without_prices(self):
        data = [
            "106/05/02",
            "45,851,963",
            "9,053,856,108",
            "--",
            "--",
            "--",
            "--",
            " 0.00",
            "15,718",
        ]
        dt = self.fetcher._make_datatuple(data)
        self.assertEqual(dt.date, datetime.datetime(2017, 5, 2))
        self.assertEqual(dt.capacity, 45851963)
        self.assertEqual(dt.turnover, 9053856108)
        self.assertEqual(dt.open, None)
        self.assertEqual(dt.high, None)
        self.assertEqual(dt.low, None)
        self.assertEqual(dt.close, None)
        self.assertEqual(dt.change, 0.0)
        self.assertEqual(dt.transaction, 15718)


class TWSEFetcerTest(unittest.TestCase, FetcherTest):
    fetcher = stock.TWSEFetcher()


class TPEXFetcherTest(unittest.TestCase, FetcherTest):
    fetcher = stock.TPEXFetcher()

    def test_make_datatuple(self):
        data = [
            "106/05/02",
            "45,851",
            "9,053,856",
            "198.50",
            "199.00",
            "195.50",
            "196.50",
            "2.00",
            "15,718",
        ]
        dt = self.fetcher._make_datatuple(data)
        self.assertEqual(dt.date, datetime.datetime(2017, 5, 2))
        self.assertEqual(dt.capacity, 45851000)
        self.assertEqual(dt.turnover, 9053856000)
        self.assertEqual(dt.open, 198.5)
        self.assertEqual(dt.high, 199.0)
        self.assertEqual(dt.low, 195.5)
        self.assertEqual(dt.close, 196.5)
        self.assertEqual(dt.change, 2.0)
        self.assertEqual(dt.transaction, 15718)

    def test_make_datatuple_without_prices(self):
        data = [
            "106/05/02",
            "45,851",
            "9,053,856",
            "--",
            "--",
            "--",
            "--",
            "0.00",
            "15,718",
        ]
        dt = self.fetcher._make_datatuple(data)
        self.assertEqual(dt.date, datetime.datetime(2017, 5, 2))
        self.assertEqual(dt.capacity, 45851000)
        self.assertEqual(dt.turnover, 9053856000)
        self.assertEqual(dt.open, None)
        self.assertEqual(dt.high, None)
        self.assertEqual(dt.low, None)
        self.assertEqual(dt.close, None)
        self.assertEqual(dt.change, 0.0)
        self.assertEqual(dt.transaction, 15718)

    def test_make_datatuple_with_asterisk(self):
        data = [
            "106/05/02＊",
            "45,851",
            "9,053,856",
            "198.50",
            "199.00",
            "195.50",
            "196.50",
            "2.00",
            "15,718",
        ]
        dt = self.fetcher._make_datatuple(data)
        self.assertEqual(dt.date, datetime.datetime(2017, 5, 2))
        self.assertEqual(dt.capacity, 45851000)
        self.assertEqual(dt.turnover, 9053856000)
        self.assertEqual(dt.open, 198.5)
        self.assertEqual(dt.high, 199.0)
        self.assertEqual(dt.low, 195.5)
        self.assertEqual(dt.close, 196.5)
        self.assertEqual(dt.change, 2.0)
        self.assertEqual(dt.transaction, 15718)


class StockTest(object):
    def test_fetch_31(self):
        self.stk.fetch_31()
        self.assertEqual(len(self.stk.data), 31)
        self.assertEqual(len(self.stk.price), 31)

    def test_date(self):
        self.assertIsInstance(self.stk.date, list)
        self.assertEqual(len(self.stk.date), len(self.stk.data))
        self.assertEqual(self.stk.date, [d.date for d in self.stk.data])

    def test_capacity(self):
        self.assertIsInstance(self.stk.capacity, list)
        self.assertEqual(len(self.stk.capacity), len(self.stk.data))
        self.assertEqual(self.stk.capacity, [d.capacity for d in self.stk.data])

    def test_turnover(self):
        self.assertIsInstance(self.stk.turnover, list)
        self.assertEqual(len(self.stk.turnover), len(self.stk.data))
        self.assertEqual(self.stk.turnover, [d.turnover for d in self.stk.data])

    def test_price(self):
        self.assertIsInstance(self.stk.price, list)
        self.assertEqual(len(self.stk.price), len(self.stk.data))
        self.assertEqual(self.stk.price, [d.close for d in self.stk.data])

    def test_high(self):
        self.assertIsInstance(self.stk.high, list)
        self.assertEqual(len(self.stk.high), len(self.stk.data))
        self.assertEqual(self.stk.high, [d.high for d in self.stk.data])

    def test_low(self):
        self.assertIsInstance(self.stk.low, list)
        self.assertEqual(len(self.stk.low), len(self.stk.data))
        self.assertEqual(self.stk.low, [d.low for d in self.stk.data])

    def test_open(self):
        self.assertIsInstance(self.stk.open, list)
        self.assertEqual(len(self.stk.open), len(self.stk.data))
        self.assertEqual(self.stk.open, [d.open for d in self.stk.data])

    def test_close(self):
        self.assertIsInstance(self.stk.close, list)
        self.assertEqual(len(self.stk.close), len(self.stk.data))
        self.assertEqual(self.stk.close, [d.close for d in self.stk.data])

    def test_change(self):
        self.assertIsInstance(self.stk.change, list)
        self.assertEqual(len(self.stk.change), len(self.stk.data))
        self.assertEqual(self.stk.change, [d.change for d in self.stk.data])

    def test_transaction(self):
        self.assertIsInstance(self.stk.transaction, list)
        self.assertEqual(len(self.stk.transaction), len(self.stk.data))
        self.assertEqual(self.stk.transaction, [d.transaction for d in self.stk.data])


class TWSEStockTest(unittest.TestCase, StockTest):
    @classmethod
    def setUpClass(cls):
        cls.stk = stock.Stock("2330")

    def test_price(self):
        self.stk.fetch(2015, 5)
        self.assertIsInstance(self.stk.price, list)
        self.assertEqual(len(self.stk.price), len(self.stk.data))
        self.assertEqual(self.stk.price, [d.close for d in self.stk.data])
        self.assertEqual(
            self.stk.price,
            [
                147.5,
                147.0,
                147.5,
                146.5,
                146.5,
                148.5,
                147.5,
                148.0,
                146.0,
                146.5,
                146.5,
                146.5,
                146.5,
                145.5,
                145.5,
                147.5,
                146.5,
                145.0,
                147.0,
                146.0,
            ],
        )

    def test_capacity(self):
        self.stk.fetch(2015, 5)
        self.assertIsInstance(self.stk.capacity, list)
        self.assertEqual(len(self.stk.capacity), len(self.stk.data))
        self.assertEqual(self.stk.capacity, [d.capacity for d in self.stk.data])
        self.assertEqual(
            self.stk.capacity,
            [
                30868640,
                27789400,
                18824208,
                21908150,
                20035646,
                20402529,
                24956498,
                19437537,
                39888654,
                24831890,
                26212375,
                26321396,
                26984912,
                41286686,
                22103852,
                16323218,
                16069726,
                24257941,
                36704395,
                61983862,
            ],
        )


class TPEXStockTest(unittest.TestCase, StockTest):
    @classmethod
    def setUpClass(cls):
        cls.stk = stock.Stock("6223")

    def test_price(self):
        self.stk.fetch(2015, 5)
        self.assertIsInstance(self.stk.price, list)
        self.assertEqual(len(self.stk.price), len(self.stk.data))
        self.assertEqual(self.stk.price, [d.close for d in self.stk.data])
        self.assertEqual(
            self.stk.price,
            [
                91.4,
                91.8,
                91.8,
                93.5,
                91.0,
                84.7,
                84.0,
                85.8,
                87.1,
                86.1,
                83.9,
                84.5,
                86.7,
                86.3,
                86.0,
                86.2,
                91.1,
                90.9,
                91.7,
                90.4,
            ],
        )

    def test_capacity(self):
        self.stk.fetch(2015, 5)
        self.assertIsInstance(self.stk.capacity, list)
        self.assertEqual(len(self.stk.capacity), len(self.stk.data))
        self.assertEqual(self.stk.capacity, [d.capacity for d in self.stk.data])
        self.assertEqual(
            self.stk.capacity,
            [
                374000,
                474000,
                468000,
                1257000,
                1079000,
                3400000,
                3424000,
                1078000,
                1433000,
                891000,
                1202000,
                1008000,
                999000,
                488000,
                706000,
                231000,
                1890000,
                782000,
                1214000,
                583000,
            ],
        )



# -*- coding: utf-8 -*-

import datetime
import json
import time
import requests
import twstock
import sys

from twstock.proxy import get_proxies

SESSION_URL = "http://mis.twse.com.tw/stock/index.jsp"
STOCKINFO_URL = (
    "http://mis.twse.com.tw/stock/api/getStockInfo.jsp?ex_ch={stock_id}&_={time}"
)

# Mock data
mock = False


def _format_stock_info(data) -> dict:
    result = {"timestamp": 0.0, "info": {}, "realtime": {}}

    # Timestamp
    result["timestamp"] = int(data["tlong"]) / 1000

    # Information
    result["info"]["code"] = data["c"]
    result["info"]["channel"] = data["ch"]
    result["info"]["name"] = data["n"]
    result["info"]["fullname"] = data["nf"]
    result["info"]["time"] = datetime.datetime.fromtimestamp(
        int(data["tlong"]) / 1000
    ).strftime("%Y-%m-%d %H:%M:%S")

    # Process best result
    def _split_best(d):
        if d:
            return d.strip("_").split("_")
        return d

    # Realtime information
    result["realtime"]["latest_trade_price"] = data.get("z", None)
    result["realtime"]["trade_volume"] = data.get("tv", None)
    result["realtime"]["accumulate_trade_volume"] = data.get("v", None)
    result["realtime"]["best_bid_price"] = _split_best(data.get("b", None))
    result["realtime"]["best_bid_volume"] = _split_best(data.get("g", None))
    result["realtime"]["best_ask_price"] = _split_best(data.get("a", None))
    result["realtime"]["best_ask_volume"] = _split_best(data.get("f", None))
    result["realtime"]["open"] = data.get("o", None)
    result["realtime"]["high"] = data.get("h", None)
    result["realtime"]["low"] = data.get("l", None)

    # Success fetching
    result["success"] = True

    return result


def _join_stock_id(stocks) -> str:
    if isinstance(stocks, list):
        return "|".join(
            [
                "{}_{}.tw".format("tse" if s in twstock.twse else "otc", s)
                for s in stocks
            ]
        )
    return "{}_{stock_id}.tw".format(
        "tse" if stocks in twstock.twse else "otc", stock_id=stocks
    )


def get_raw(stocks) -> dict:
    req = requests.Session()
    req.get(SESSION_URL, proxies=get_proxies())

    r = req.get(
        STOCKINFO_URL.format(
            stock_id=_join_stock_id(stocks), time=int(time.time()) * 1000
        )
    )

    if sys.version_info < (3, 5):
        try:
            return r.json()
        except ValueError:
            return {"rtmessage": "json decode error", "rtcode": "5000"}
    else:
        try:
            return r.json()
        except json.decoder.JSONDecodeError:
            return {"rtmessage": "json decode error", "rtcode": "5000"}


def get(stocks, retry=3):
    # Prepare data
    data = get_raw(stocks) if not mock else twstock.mock.get(stocks)

    # Set success
    data["success"] = False

    # JSONdecode error, could be too fast, retry
    if data["rtcode"] == "5000":
        # XXX: Stupit retry, you will dead here
        if retry:
            return get(stocks, retry - 1)
        return data

    # No msgArray, dead
    if "msgArray" not in data:
        return data

    # Check have data
    if not len(data["msgArray"]):
        data["rtmessage"] = "Empty Query."
        data["rtcode"] = "5001"
        return data

    # Return multiple stock data
    if isinstance(stocks, list):
        result = {
            data["info"]["code"]: data
            for data in map(_format_stock_info, data["msgArray"])
        }
        result["success"] = True
        return result

    return _format_stock_info(data["msgArray"][0])



import unittest
from twstock import stock
from twstock import analytics
from twstock import legacy


class AnalyticsTest(unittest.TestCase):
    def setUp(self):
        self.legacy = legacy.LegacyAnalytics()
        self.ng = analytics.Analytics()

    def test_continuous(self):
        data = [1, 2, 3, 4, 5, 6, 7]
        legacy_result = self.legacy.cal_continue(data)
        ng_result = self.ng.continuous(data)
        self.assertEqual(ng_result, legacy_result)
        self.assertEqual(ng_result, 6)

        data = [1, 2, 3, 4, 1, 2]
        legacy_result = self.legacy.cal_continue(data)
        ng_result = self.ng.continuous(data)
        self.assertEqual(ng_result, legacy_result)
        self.assertEqual(ng_result, 1)

        data = [1, 2, 3, 4, 1, 2, 3, 4, 5]
        legacy_result = self.legacy.cal_continue(data)
        ng_result = self.ng.continuous(data)
        self.assertEqual(ng_result, legacy_result)
        self.assertEqual(ng_result, 4)

        data = [5, 4, 3, 2, 1]
        legacy_result = self.legacy.cal_continue(data)
        ng_result = self.ng.continuous(data)
        self.assertEqual(ng_result, legacy_result)
        self.assertEqual(ng_result, -4)

        data = [5, 4, 3, 2, 1, 5, 4, 3]
        legacy_result = self.legacy.cal_continue(data)
        ng_result = self.ng.continuous(data)
        self.assertEqual(ng_result, legacy_result)
        self.assertEqual(ng_result, -2)

    def test_moving_average(self):
        data = [50, 60, 70, 75]

        # Legacy moving_average will affect data argument's data
        ng_result = self.ng.moving_average(data, 2)
        legacy_result = self.legacy.moving_average(data, 2)
        self.assertEqual(ng_result, legacy_result)
        self.assertEqual(ng_result, [55.0, 65.0, 72.5])

    def test_ma_bias_ratio(self):
        data = [50, 60, 70, 75, 80, 88, 102, 105, 106]
        self.ng.price = data
        ng_result = self.ng.ma_bias_ratio(3, 6)
        legacy_result = self.legacy.ma_bias_ratio(3, 6, data)
        self.assertEqual(ng_result, legacy_result)

        data = [75, 72, 77, 85, 100, 65, 60, 55, 52, 45]
        self.ng.price = data
        ng_result = self.ng.ma_bias_ratio(3, 6)
        legacy_result = self.legacy.ma_bias_ratio(3, 6, data)
        self.assertEqual(ng_result, legacy_result)

    def test_ma_bias_ratio_pivot(self):
        data = [50, 60, 70, 75, 80, 88, 102, 105, 106]
        legacy_result = self.legacy.ma_bias_ratio_point(data, 5, False)
        ng_result = self.ng.ma_bias_ratio_pivot(data, 5, False)
        self.assertEqual(legacy_result, ng_result)

        legacy_result = self.legacy.ma_bias_ratio_point(data, 5, True)
        ng_result = self.ng.ma_bias_ratio_pivot(data, 5, True)
        self.assertEqual(legacy_result, ng_result)

        data = [75, 72, 77, 85, 100, 65, 60, 55, 52, 45]
        legacy_result = self.legacy.ma_bias_ratio_point(data, 5, False)
        ng_result = self.ng.ma_bias_ratio_pivot(data, 5, False)
        self.assertEqual(legacy_result, ng_result)

        legacy_result = self.legacy.ma_bias_ratio_point(data, 5, True)
        ng_result = self.ng.ma_bias_ratio_pivot(data, 5, True)
        self.assertEqual(legacy_result, ng_result)


class BestFourPointTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.stock = stock.Stock("2330")
        self.stock.fetch(2015, 5)
        self.legacy = legacy.LegacyBestFourPoint(self.stock)
        self.ng = analytics.BestFourPoint(self.stock)

    def test_bias_ratio(self):
        self.assertEqual(self.ng.bias_ratio(), self.legacy.bias_ratio())
        self.assertEqual(self.ng.bias_ratio(True), self.legacy.bias_ratio(True))

    def test_best_buy_1(self):
        self.assertEqual(self.ng.best_buy_1(), self.legacy.best_buy_1())

    def test_best_buy_2(self):
        self.assertEqual(self.ng.best_buy_2(), self.legacy.best_buy_2())

    def test_best_buy_3(self):
        self.assertEqual(self.ng.best_buy_3(), self.legacy.best_buy_3())

    def test_best_buy_4(self):
        self.assertEqual(self.ng.best_buy_4(), self.legacy.best_buy_4())

    def test_best_sell_1(self):
        self.assertEqual(self.ng.best_sell_1(), self.legacy.best_sell_1())

    def test_best_sell_2(self):
        self.assertEqual(self.ng.best_sell_2(), self.legacy.best_sell_2())

    def test_best_sell_3(self):
        self.assertEqual(self.ng.best_sell_3(), self.legacy.best_sell_3())

    def test_best_sell_4(self):
        self.assertEqual(self.ng.best_sell_4(), self.legacy.best_sell_4())

    def test_best_four_point_to_buy(self):
        self.assertEqual(
            self.ng.best_four_point_to_buy(), self.legacy.best_four_point_to_buy()
        )

    def test_best_four_point_to_sell(self):
        self.assertEqual(
            self.ng.best_four_point_to_sell(), self.legacy.best_four_point_to_sell()
        )

    def test_best_four_point(self):
        self.stock.fetch(2014, 5)
        self.assertEqual(self.ng.best_four_point(), self.legacy.best_four_point())

        self.stock.fetch(2015, 5)
        self.assertEqual(self.ng.best_four_point(), self.legacy.best_four_point())

        self.stock.fetch(2016, 5)
        self.assertEqual(self.ng.best_four_point(), self.legacy.best_four_point())

        self.stock.fetch(2017, 5)
        self.assertEqual(self.ng.best_four_point(), self.legacy.best_four_point())



# -*- coding: utf-8 -*-
import argparse
from twstock.codes import __update_codes
from twstock.cli import best_four_point
from twstock.cli import stock
from twstock.cli import realtime


def run():
    parser = argparse.ArgumentParser()

    parser.add_argument("-b", "--bfp", nargs="+")
    parser.add_argument("-s", "--stock", nargs="+")
    parser.add_argument("-r", "--realtime", nargs="+")
    parser.add_argument(
        "-U", "--upgrade-codes", action="store_true", help="Update entites codes"
    )
    args = parser.parse_args()

    if args.bfp:
        best_four_point.run(args.bfp)
    elif args.stock:
        stock.run(args.stock)
    elif args.realtime:
        realtime.run(args.realtime)
    elif args.upgrade_codes:
        print("Start to update codes")
        __update_codes()
        print("Done!")
    else:
        parser.print_help()



# -*- coding: utf-8 -*-


class Analytics(object):
    def continuous(self, data):
        diff = [1 if data[-i] > data[-i - 1] else -1 for i in range(1, len(data))]
        cont = 0
        for v in diff:
            if v == diff[0]:
                cont += 1
            else:
                break
        return cont * diff[0]

    def moving_average(self, data, days):
        result = []
        data = data[:]
        for _ in range(len(data) - days + 1):
            result.append(round(sum(data[-days:]) / days, 2))
            data.pop()
        return result[::-1]

    def ma_bias_ratio(self, day1, day2):
        """Calculate moving average bias ratio"""
        data1 = self.moving_average(self.price, day1)
        data2 = self.moving_average(self.price, day2)
        result = [
            data1[-i] - data2[-i] for i in range(1, min(len(data1), len(data2)) + 1)
        ]

        return result[::-1]

    def ma_bias_ratio_pivot(self, data, sample_size=5, position=False):
        """Calculate pivot point"""
        sample = data[-sample_size:]

        if position is True:
            check_value = max(sample)
            pre_check_value = max(sample) > 0
        elif position is False:
            check_value = min(sample)
            pre_check_value = max(sample) < 0

        return (
            (
                sample_size - sample.index(check_value) < 4
                and sample.index(check_value) != sample_size - 1
                and pre_check_value
            ),
            sample_size - sample.index(check_value) - 1,
            check_value,
        )


class BestFourPoint(object):
    BEST_BUY_WHY = ["量大收紅", "量縮價不跌", "三日均價由下往上", "三日均價大於六日均價"]
    BEST_SELL_WHY = ["量大收黑", "量縮價跌", "三日均價由上往下", "三日均價小於六日均價"]

    def __init__(self, stock):
        self.stock = stock

    def bias_ratio(self, position=False):
        return self.stock.ma_bias_ratio_pivot(
            self.stock.ma_bias_ratio(3, 6), position=position
        )

    def plus_bias_ratio(self):
        return self.bias_ratio(True)

    def mins_bias_ratio(self):
        return self.bias_ratio(False)

    def best_buy_1(self):
        return (
            self.stock.capacity[-1] > self.stock.capacity[-2]
            and self.stock.price[-1] > self.stock.open[-1]
        )

    def best_buy_2(self):
        return (
            self.stock.capacity[-1] < self.stock.capacity[-2]
            and self.stock.price[-1] > self.stock.open[-2]
        )

    def best_buy_3(self):
        return (
            self.stock.continuous(self.stock.moving_average(self.stock.price, 3)) == 1
        )

    def best_buy_4(self):
        return (
            self.stock.moving_average(self.stock.price, 3)[-1]
            > self.stock.moving_average(self.stock.price, 6)[-1]
        )

    def best_sell_1(self):
        return (
            self.stock.capacity[-1] > self.stock.capacity[-2]
            and self.stock.price[-1] < self.stock.open[-1]
        )

    def best_sell_2(self):
        return (
            self.stock.capacity[-1] < self.stock.capacity[-2]
            and self.stock.price[-1] < self.stock.open[-2]
        )

    def best_sell_3(self):
        return (
            self.stock.continuous(self.stock.moving_average(self.stock.price, 3)) == -1
        )

    def best_sell_4(self):
        return (
            self.stock.moving_average(self.stock.price, 3)[-1]
            < self.stock.moving_average(self.stock.price, 6)[-1]
        )

    def best_four_point_to_buy(self):
        result = []
        check = [
            self.best_buy_1(),
            self.best_buy_2(),
            self.best_buy_3(),
            self.best_buy_4(),
        ]
        if self.mins_bias_ratio() and any(check):
            for index, v in enumerate(check):
                if v:
                    result.append(self.BEST_BUY_WHY[index])
        else:
            return False
        return ", ".join(result)

    def best_four_point_to_sell(self):
        result = []
        check = [
            self.best_sell_1(),
            self.best_sell_2(),
            self.best_sell_3(),
            self.best_sell_4(),
        ]
        if self.plus_bias_ratio() and any(check):
            for index, v in enumerate(check):
                if v:
                    result.append(self.BEST_SELL_WHY[index])
        else:
            return False
        return ", ".join(result)

    def best_four_point(self):
        buy = self.best_four_point_to_buy()
        sell = self.best_four_point_to_sell()
        if buy:
            return (True, buy)
        elif sell:
            return (False, sell)

        return None



"""Taiwan Stock Opendata with realtime - twstock"""

import importlib.metadata

from twstock import stock
from twstock import analytics
from twstock import cli
from twstock import mock
from twstock import realtime

from twstock.analytics import BestFourPoint
from twstock.codes import __update_codes, twse, tpex, codes
from twstock.stock import Stock

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "1.4.0"
